# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .statement_type import StatementType
import pydantic
import typing_extensions
from ..core.serialization import FieldMetadata
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Statement(UniversalBaseModel):
    id: str
    statement: str
    type: typing.Optional[StatementType] = pydantic.Field(default=None)
    """
    The type/category of the statement.
    """

    is_formalized: typing.Optional[bool] = None
    formalization: typing.Optional[typing.List[typing.Optional[typing.Any]]] = None
    z_3_formalization: typing_extensions.Annotated[typing.Optional[str], FieldMetadata(alias="z3_formalization")] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.formalize_circuit_response import FormalizeCircuitResponse
from ...core.pydantic_utilities import parse_obj_as
from ...errors.unprocessable_entity_error import UnprocessableEntityError
from ...types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...types.generate_code_response import GenerateCodeResponse
from ...types.refine_code_response import RefineCodeResponse
from ...types.netlist import Netlist
from ...types.statement import Statement
from ...types.mapping import Mapping
from ...types.optimize_netlist_response import OptimizeNetlistResponse
from ...core.serialization import convert_and_respect_annotation_metadata
from ...types.verify_circuit_code_response import VerifyCircuitCodeResponse
from ...core.client_wrapper import AsyncClientWrapper

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class CircuitClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def formalize(
        self, *, query: str, request_options: typing.Optional[RequestOptions] = None
    ) -> FormalizeCircuitResponse:
        """
        Formalize a query about a circuit into a dictionary of constraints

        Parameters
        ----------
        query : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FormalizeCircuitResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.circuit.formalize(
            query="query",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/query/formalize",
            method="POST",
            json={
                "query": query,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    FormalizeCircuitResponse,
                    parse_obj_as(
                        type_=FormalizeCircuitResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def generate(self, *, query: str, request_options: typing.Optional[RequestOptions] = None) -> GenerateCodeResponse:
        """
        Generate GDS factory code to create a circuit

        Parameters
        ----------
        query : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        GenerateCodeResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.circuit.generate(
            query="query",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/generate",
            method="POST",
            json={
                "query": query,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    GenerateCodeResponse,
                    parse_obj_as(
                        type_=GenerateCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def refine(
        self,
        *,
        query: str,
        feedback: typing.Optional[str] = OMIT,
        code: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RefineCodeResponse:
        """
        Refine GDS factory code to create a circuit

        Parameters
        ----------
        query : str

        feedback : typing.Optional[str]

        code : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RefineCodeResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.circuit.refine(
            query="query",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/refine",
            method="POST",
            json={
                "query": query,
                "feedback": feedback,
                "code": code,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    RefineCodeResponse,
                    parse_obj_as(
                        type_=RefineCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def optimize(
        self,
        *,
        netlist: Netlist,
        statements: typing.Sequence[Statement],
        mappings: typing.Sequence[Mapping],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizeNetlistResponse:
        """
        Optimize a PIC circuit with given cost and constraints

        Parameters
        ----------
        netlist : Netlist

        statements : typing.Sequence[Statement]

        mappings : typing.Sequence[Mapping]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizeNetlistResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic, Mapping, Netlist, PicComponent, Statement

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.circuit.optimize(
            netlist=Netlist(
                name="name",
                instances={
                    "key": PicComponent(
                        component="component",
                    )
                },
                connections={"key": "value"},
                ports={"key": "value"},
            ),
            statements=[
                Statement(
                    id="id",
                    statement="statement",
                )
            ],
            mappings=[
                Mapping(
                    statement_id="statement_id",
                    expression_idx=1,
                    variable="variable",
                    computation_name="computation_name",
                    computation_arguments={"key": 1.1},
                )
            ],
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/optimize",
            method="POST",
            json={
                "netlist": convert_and_respect_annotation_metadata(
                    object_=netlist, annotation=Netlist, direction="write"
                ),
                "statements": convert_and_respect_annotation_metadata(
                    object_=statements, annotation=typing.Sequence[Statement], direction="write"
                ),
                "mappings": convert_and_respect_annotation_metadata(
                    object_=mappings, annotation=typing.Sequence[Mapping], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    OptimizeNetlistResponse,
                    parse_obj_as(
                        type_=OptimizeNetlistResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def verify(
        self, *, code: str, request_options: typing.Optional[RequestOptions] = None
    ) -> VerifyCircuitCodeResponse:
        """
        Verifies that the code for a circuit

        Parameters
        ----------
        code : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        VerifyCircuitCodeResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.circuit.verify(
            code="code",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/verifycode",
            method="POST",
            json={
                "code": code,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    VerifyCircuitCodeResponse,
                    parse_obj_as(
                        type_=VerifyCircuitCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncCircuitClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def formalize(
        self, *, query: str, request_options: typing.Optional[RequestOptions] = None
    ) -> FormalizeCircuitResponse:
        """
        Formalize a query about a circuit into a dictionary of constraints

        Parameters
        ----------
        query : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FormalizeCircuitResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.circuit.formalize(
                query="query",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/query/formalize",
            method="POST",
            json={
                "query": query,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    FormalizeCircuitResponse,
                    parse_obj_as(
                        type_=FormalizeCircuitResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def generate(
        self, *, query: str, request_options: typing.Optional[RequestOptions] = None
    ) -> GenerateCodeResponse:
        """
        Generate GDS factory code to create a circuit

        Parameters
        ----------
        query : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        GenerateCodeResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.circuit.generate(
                query="query",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/generate",
            method="POST",
            json={
                "query": query,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    GenerateCodeResponse,
                    parse_obj_as(
                        type_=GenerateCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def refine(
        self,
        *,
        query: str,
        feedback: typing.Optional[str] = OMIT,
        code: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RefineCodeResponse:
        """
        Refine GDS factory code to create a circuit

        Parameters
        ----------
        query : str

        feedback : typing.Optional[str]

        code : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RefineCodeResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.circuit.refine(
                query="query",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/refine",
            method="POST",
            json={
                "query": query,
                "feedback": feedback,
                "code": code,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    RefineCodeResponse,
                    parse_obj_as(
                        type_=RefineCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def optimize(
        self,
        *,
        netlist: Netlist,
        statements: typing.Sequence[Statement],
        mappings: typing.Sequence[Mapping],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizeNetlistResponse:
        """
        Optimize a PIC circuit with given cost and constraints

        Parameters
        ----------
        netlist : Netlist

        statements : typing.Sequence[Statement]

        mappings : typing.Sequence[Mapping]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizeNetlistResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic, Mapping, Netlist, PicComponent, Statement

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.circuit.optimize(
                netlist=Netlist(
                    name="name",
                    instances={
                        "key": PicComponent(
                            component="component",
                        )
                    },
                    connections={"key": "value"},
                    ports={"key": "value"},
                ),
                statements=[
                    Statement(
                        id="id",
                        statement="statement",
                    )
                ],
                mappings=[
                    Mapping(
                        statement_id="statement_id",
                        expression_idx=1,
                        variable="variable",
                        computation_name="computation_name",
                        computation_arguments={"key": 1.1},
                    )
                ],
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/optimize",
            method="POST",
            json={
                "netlist": convert_and_respect_annotation_metadata(
                    object_=netlist, annotation=Netlist, direction="write"
                ),
                "statements": convert_and_respect_annotation_metadata(
                    object_=statements, annotation=typing.Sequence[Statement], direction="write"
                ),
                "mappings": convert_and_respect_annotation_metadata(
                    object_=mappings, annotation=typing.Sequence[Mapping], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    OptimizeNetlistResponse,
                    parse_obj_as(
                        type_=OptimizeNetlistResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def verify(
        self, *, code: str, request_options: typing.Optional[RequestOptions] = None
    ) -> VerifyCircuitCodeResponse:
        """
        Verifies that the code for a circuit

        Parameters
        ----------
        code : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        VerifyCircuitCodeResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.circuit.verify(
                code="code",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/verifycode",
            method="POST",
            json={
                "code": code,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    VerifyCircuitCodeResponse,
                    parse_obj_as(
                        type_=VerifyCircuitCodeResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



import ipywidgets as widgets # type: ignore
from IPython.display import display, Math, HTML # type: ignore
from dataclasses import dataclass, field
import hypernetx as hnx # type: ignore
import matplotlib.pyplot as plt # type: ignore
import re

OPTION_LIST = {
    "Select a template": [],
    "IMAGING TELESCOPE": [
        "Resolution (panchromatic)",
        "Ground sampling distance (panchromatic)",
        "Resolution (multispectral)",
        "Ground sampling distance (multispectral)",
        "Altitude",
        "Half field of view",
        "Mirror aperture",
        "F-number",
        "Focal length",
        "Pixel size (panchromatic)",
        "Pixel size (multispectral)",
        "Swath width",
    ],
    "PAYLOAD": [
        "Resolution (panchromatic)",
        "Ground sampling distance (panchromatic)",
        "Resolution (multispectral)",
        "Ground sampling distance (multispectral)",
        "Altitude",
        "Half field of view",
        "Mirror aperture",
        "F-number",
        "Focal length",
        "Pixel size (panchromatic)",
        "Swath width",
    ],
}

IMAGING_TELESCOPE = {
    "Resolution (panchromatic)": 1.23529,
    "Ground sampling distance (panchromatic)": 0.61765,
    "Resolution (multispectral)": 1.81176,
    "Ground sampling distance (multispectral)": 0.90588,
    "Altitude": 420000,
    "Half field of view": 0.017104227,
    "Mirror aperture": 0.85,
    "F-number": 6.0,
    "Focal length": 5.1,
    "Pixel size (panchromatic)": 7.5e-6,
    "Pixel size (multispectral)": 11e-6,
    "Swath width": 14368.95,
}

IMAGING_TELESCOPE_UNITS = {
    "Resolution (panchromatic)": "m",
    "Ground sampling distance (panchromatic)": "m",
    "Resolution (multispectral)": "m",
    "Ground sampling distance (multispectral)": "m",
    "Altitude": "m",
    "Half field of view": "rad",
    "Mirror aperture": "m",
    "F-number": "dimensionless",
    "Focal length": "m",
    "Pixel size (panchromatic)": "m",
    "Pixel size (multispectral)": "m",
    "Swath width": "m",
}

PAYLOAD_1 = {
    "Resolution (panchromatic)": 15.4,
    "Ground sampling distance (panchromatic)": 7.7,
    "Resolution (multispectral)": 0.0,
    "Ground sampling distance (multispectral)": 0.,
    "Altitude": 420000,
    "Half field of view": 0.005061455,
    "Mirror aperture": 0.85,
    "F-number": 1.,
    "Focal length": 0.3,
    "Pixel size (panchromatic)": 5.5e-6,
    "Swath width": 4251.66,
}


@dataclass
class Requirement:
    requirement_name: str
    latex_symbol: str
    value: int
    units: str
    tolerance: float
    sympy_symbol: str = field(init=False)

    def __post_init__(self):
        self.sympy_symbol = self.latex_symbol.replace("{", "").replace("}", "")

    @property
    def is_fixed(self):
        return self.tolerance == 0.0


def _find_symbol(name, variable_dict):

    matching_keys = [
        key for key, value in variable_dict.items() if name in value["name"]
    ]

    if not matching_keys:
        matching_keys.append("unknown")

    return matching_keys[0]


def requirements_from_table(results, variable_dict):
    requirements = []

    for key, value in results["values"].items():

        latex_symbol = _find_symbol(key, variable_dict)

        name = key
        numerical_value = value["Value"]
        unit = value["Units"]

        requirements.append(
            Requirement(
                requirement_name=name,
                latex_symbol=latex_symbol,
                value=numerical_value,
                units=unit,
                tolerance=0.0,
            )
        )

    return requirements


def interactive_table(preset_options_dict, variable_dict):
    """
    Creates an interactive table with a dropdown for selecting options.

    Parameters:
    options_dict (dict): A dictionary where keys are dropdown options and
      values are lists of row names.

    Returns:
    dict: A dictionary containing user inputs for the selected rows.
    """

    variable_names = [details["name"] for details in variable_dict.values()]

    # Placeholder for result dictionary
    result = {}

    # Create dropdown for options
    dropdown = widgets.Dropdown(
        options=list(preset_options_dict.keys()),
        description="Select Option:",
        style={"description_width": "initial"},
    )

    # Dictionary to hold widgets for user input
    value_widgets = {}

    # VBox to stack rows vertically
    rows_output = widgets.VBox()

    # Output widget for confirmation messages
    message_output = widgets.Output()

    # Mutable container to store the current name label width
    name_label_width = ["150px"]  # Default width

    # Function to display the table based on the dropdown selection
    def display_table(change):
        selected_option = change["new"]

        # Clear existing rows
        rows_output.children = []
        value_widgets.clear()

        if selected_option in preset_options_dict:
            rows = preset_options_dict[selected_option]

            if selected_option != "Select a template":
                max_name_length = max(len(name) for name in rows)
                # Update the name_label_width based on the longest row name
                name_label_width[0] = f"{max_name_length + 2}ch"
            else:
                max_name_length = 40
                # Update the name_label_width based on the longest row name
                name_label_width[0] = f"{max_name_length + 2}ch"

            # Add Headers
            header_labels = [
                widgets.Label(
                    value="Name",
                    layout=widgets.Layout(width=name_label_width[0]),
                    style={'font_weight': 'bold'}
                ),
                widgets.Label(
                    value="Value",
                    layout=widgets.Layout(width="150px"),
                    style={'font_weight': 'bold'}
                ),
                widgets.Label(
                    value="Units",
                    layout=widgets.Layout(width="150px"),
                    style={'font_weight': 'bold'}
                ),
            ]

            # Combine header labels into a horizontal box
            header = widgets.HBox(header_labels)
            header.layout = widgets.Layout(
                border='1px solid black',
                padding='5px',
            )

            # Add the header to the rows_output VBox
            rows_output.children += (header,)

            for row_name in rows:
                # Create name label with dynamic width
                name_label = widgets.Label(
                    value=row_name,
                    layout=widgets.Layout(width=name_label_width[0]),
                )

                # Depending on the selected option, set default values
                if selected_option == "IMAGING TELESCOPE":
                    default_value = IMAGING_TELESCOPE.get(row_name, 0.0)
                    default_unit = IMAGING_TELESCOPE_UNITS.get(row_name, "")
                # elif selected_option == "LIDAR":
                #     default_value = LIDAR.get(row_name, 0.0)
                elif selected_option == "PAYLOAD":
                    default_value = PAYLOAD_1.get(row_name, 0.0)
                    default_unit = IMAGING_TELESCOPE_UNITS.get(row_name, "")
                else:
                    default_value = 0.0
                    default_unit = ""

                # Create input widgets
                value_text = widgets.FloatText(
                    value=default_value,
                    layout=widgets.Layout(width="150px"),
                )
                units_text = widgets.Text(
                    layout=widgets.Layout(width="150px"),
                    value=default_unit
                )

                # Combine widgets into a horizontal box
                row = widgets.HBox(
                    [
                        name_label,
                        value_text,
                        units_text,
                    ]
                )

                # Store the row widgets
                value_widgets[row_name] = row

                # Add the row to the rows_output VBox
                rows_output.children += (row,)

    # Attach handler to dropdown
    dropdown.observe(display_table, names="value")
    display(dropdown)
    display(rows_output)
    display(message_output)

    # Function to collect and store user inputs
    def submit_values(_):
        updated_values = {}

        for key, widget in value_widgets.items():
            variable = widget.children[0].value
            if key.startswith("req_"):
                updated_values[variable] = {
                    "Value": widget.children[1].value,
                    "Units": widget.children[2].value,
                }
            else:
                updated_values[key] = {
                    "Value": widget.children[1].value,
                    "Units": widget.children[2].value,
                }

        result["values"] = updated_values

        # Display confirmation message
        with message_output:
            message_output.clear_output()

    # Function to add a new requirement row
    def add_req(_):

        unique_key = (
            f"req_{len([k for k in value_widgets if k.startswith('req_')]) + 1}"
        )

        # Create a dropdown for variable selection with dynamic width
        variable_dropdown = widgets.Dropdown(
            options=variable_names,
            description="Variable:",
            style={"description_width": "initial"},
            layout=widgets.Layout(width=2 * name_label_width[0]),
        )
        value_text = widgets.FloatText(
            placeholder="Value",
            layout=widgets.Layout(width="150px"),
        )

        units_text = widgets.Text(
            placeholder="Units", layout=widgets.Layout(width="150px")
        )

        new_row = widgets.HBox(
            [variable_dropdown, value_text, units_text]
        )

        rows_output.children += (new_row,)
        value_widgets[unique_key] = new_row

    submit_button = widgets.Button(description="Submit")
    submit_button.on_click(submit_values)

    add_req_button = widgets.Button(description="Add Requirement")
    add_req_button.on_click(add_req)

    buttons_box = widgets.HBox([submit_button, add_req_button])
    display(buttons_box)

    return result


def display_formatted_answers(equations_dict):
    """
    Display LaTeX formatted equations and numerical results from a nested 
    dictionary structure in Jupyter Notebook.

    Parameters:
    equations_dict (dict): The dictionary containing the equations.
    """
    results = equations_dict.get('results', {})
    print("We identified the following equations that are relevant to your requirements:")

    for key, value in results.items():
        latex_equation = value.get('latex_equation')
        lhs = value.get('lhs')
        rhs = value.get('rhs')
        match = value.get('match')
        if latex_equation:
            display(Math(latex_equation))
            print(f"For provided values:\nleft hand side = {lhs}\nright hand side = {rhs}")
            if match:
                print("Provided requirements fulfill this mathematical relation")
        else:
            print(f"No LaTeX equation found for {key}")


def display_results(equations_dict):

    results = equations_dict.get('results', {})
    not_match_counter = 0

    for key, value in results.items():
        match = value.get('match')
        latex_equation = value.get('latex_equation')
        lhs = value.get('lhs')
        rhs = value.get('rhs')
        if not match:
            not_match_counter += 1
            display(HTML(
                '<p style="color:red; '
                'font-weight:bold; '
                'font-family:\'Times New Roman\'; '
                'font-size:16px;">'
                'Provided requirements DO NOT fulfill the following mathematical relation:'
                '</p>'
                ))           
            display(Math(latex_equation))
            print(f"For provided values:\nleft hand side = {lhs}\nright hand side = {rhs}")
    if not_match_counter == 0:
        display(HTML(
            '<p style="color:green; '
            'font-weight:bold; '
            'font-family:\'Times New Roman\'; '
            'font-size:16px;">'
            'Requirements you provided do not cause any conflicts'
            '</p>'
        ))


def _get_latex_string_format(input_string):
    """
    Properly formats LaTeX strings for matplotlib when text.usetex is False.
    No escaping needed since mathtext handles backslashes properly.
    """
    return f"${input_string}$"  # No backslash escaping required


def _get_requirements_set(requirements):
    variable_set = set()
    for req in requirements:
        variable_set.add(req['latex_symbol'])

    return variable_set


def _find_vars_in_eq(equation, variable_set):
    patterns = [re.escape(var) for var in variable_set]
    combined_pattern = r'|'.join(patterns)
    matches = re.findall(combined_pattern, equation)
    return {fr"${match}$" for match in matches}


def _add_used_vars_to_results(api_results, api_requirements):
    requirements = _get_requirements_set(api_requirements)

    for key, value in api_results['results'].items():
        latex_equation = value.get('latex_equation')
        # print(latex_equation)
        if latex_equation:
            used_vars = _find_vars_in_eq(latex_equation, requirements)
            api_results['results'][key]['used_vars'] = used_vars

    return api_results


def get_eq_hypergraph(api_results, api_requirements):
    # Disable external LaTeX rendering, using matplotlib's mathtext instead
    plt.rcParams['text.usetex'] = False
    plt.rcParams['mathtext.fontset'] = 'stix'
    plt.rcParams['font.family'] = 'serif'

    api_results = _add_used_vars_to_results(api_results, api_requirements)

    # Prepare the data for HyperNetX visualization
    hyperedges = {}
    for eq, details in api_results["results"].items():
        hyperedges[_get_latex_string_format(
            details["latex_equation"])] = details["used_vars"]

    # Create the hypergraph using HyperNetX
    H = hnx.Hypergraph(hyperedges)

    # Plot the hypergraph with enhanced clarity
    plt.figure(figsize=(16, 12))

    # Draw the hypergraph with node and edge labels
    hnx.draw(
        H, 
        with_edge_labels=True, 
        edge_labels_on_edge=False,
        node_labels_kwargs={'fontsize': 14}, 
        edge_labels_kwargs={'fontsize': 14},
        layout_kwargs={'seed': 42, 'scale': 2.5}  
    )

    node_labels = list(H.nodes)
    symbol_explanations = _get_node_names_for_node_lables(node_labels, api_requirements)

    # Adding the symbol explanations as a legend
    explanation_text = "\n".join([f"${symbol}$: {desc}" for symbol, desc in symbol_explanations])
    plt.annotate(
        explanation_text, 
        xy=(1.05, 0.5), 
        xycoords='axes fraction', 
        fontsize=14, 
        verticalalignment='center'
    )

    plt.title(r"Enhanced Hypergraph of Equations and Variables", fontsize=20)
    plt.show()


def _get_node_names_for_node_lables(node_labels, api_requirements):

    # Create the output list
    node_names = []

    # Iterate through each symbol in S
    for symbol in node_labels:
        # Search for the matching requirement
        symbol = symbol.replace("$", "")
        for req in api_requirements:
            if req['latex_symbol'] == symbol:
                # Add the matching tuple to SS
                node_names.append((req["latex_symbol"], req["requirement_name"]))
                break  # Stop searching once a match is found

    return node_names



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .pic_component_info_value import PicComponentInfoValue
from .pic_component_settings_value import PicComponentSettingsValue
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class PicComponent(UniversalBaseModel):
    component: str
    info: typing.Optional[typing.Dict[str, PicComponentInfoValue]] = None
    settings: typing.Optional[typing.Dict[str, PicComponentSettingsValue]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .environment import AxiomaticEnvironment
import os
import httpx
from .core.api_error import ApiError
from .core.client_wrapper import SyncClientWrapper
from .requirements.client import RequirementsClient
from .pic.client import PicClient
from .lean.client import LeanClient
from .experimental.client import ExperimentalClient
from .formalization.client import FormalizationClient
from .generic.client import GenericClient
from .document.client import DocumentClient
from .code_execution.client import CodeExecutionClient
from .core.request_options import RequestOptions
from .core.pydantic_utilities import parse_obj_as
from json.decoder import JSONDecodeError
from .core.client_wrapper import AsyncClientWrapper
from .requirements.client import AsyncRequirementsClient
from .pic.client import AsyncPicClient
from .lean.client import AsyncLeanClient
from .experimental.client import AsyncExperimentalClient
from .formalization.client import AsyncFormalizationClient
from .generic.client import AsyncGenericClient
from .document.client import AsyncDocumentClient
from .code_execution.client import AsyncCodeExecutionClient


class Axiomatic:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : AxiomaticEnvironment
        The environment to use for requests from the client. from .environment import AxiomaticEnvironment



        Defaults to AxiomaticEnvironment.DEFAULT



    api_key : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.Client]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from axiomatic import Axiomatic

    client = Axiomatic(
        api_key="YOUR_API_KEY",
    )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: AxiomaticEnvironment = AxiomaticEnvironment.DEFAULT,
        api_key: typing.Optional[str] = os.getenv("AXIOMATIC_API_KEY"),
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.Client] = None,
    ):
        _defaulted_timeout = timeout if timeout is not None else 60 if httpx_client is None else None
        if api_key is None:
            raise ApiError(
                body="The client must be instantiated be either passing in api_key or setting AXIOMATIC_API_KEY"
            )
        self._client_wrapper = SyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.Client(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.Client(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self.requirements = RequirementsClient(client_wrapper=self._client_wrapper)
        self.pic = PicClient(client_wrapper=self._client_wrapper)
        self.lean = LeanClient(client_wrapper=self._client_wrapper)
        self.experimental = ExperimentalClient(client_wrapper=self._client_wrapper)
        self.formalization = FormalizationClient(client_wrapper=self._client_wrapper)
        self.generic = GenericClient(client_wrapper=self._client_wrapper)
        self.document = DocumentClient(client_wrapper=self._client_wrapper)
        self.code_execution = CodeExecutionClient(client_wrapper=self._client_wrapper)

    def trigger_error_sentry_debug_get(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.trigger_error_sentry_debug_get()
        """
        _response = self._client_wrapper.httpx_client.request(
            "sentry-debug",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def health_check_health_check_get(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.health_check_health_check_get()
        """
        _response = self._client_wrapper.httpx_client.request(
            "health_check",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncAxiomatic:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : AxiomaticEnvironment
        The environment to use for requests from the client. from .environment import AxiomaticEnvironment



        Defaults to AxiomaticEnvironment.DEFAULT



    api_key : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.AsyncClient]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from axiomatic import AsyncAxiomatic

    client = AsyncAxiomatic(
        api_key="YOUR_API_KEY",
    )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: AxiomaticEnvironment = AxiomaticEnvironment.DEFAULT,
        api_key: typing.Optional[str] = os.getenv("AXIOMATIC_API_KEY"),
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.AsyncClient] = None,
    ):
        _defaulted_timeout = timeout if timeout is not None else 60 if httpx_client is None else None
        if api_key is None:
            raise ApiError(
                body="The client must be instantiated be either passing in api_key or setting AXIOMATIC_API_KEY"
            )
        self._client_wrapper = AsyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self.requirements = AsyncRequirementsClient(client_wrapper=self._client_wrapper)
        self.pic = AsyncPicClient(client_wrapper=self._client_wrapper)
        self.lean = AsyncLeanClient(client_wrapper=self._client_wrapper)
        self.experimental = AsyncExperimentalClient(client_wrapper=self._client_wrapper)
        self.formalization = AsyncFormalizationClient(client_wrapper=self._client_wrapper)
        self.generic = AsyncGenericClient(client_wrapper=self._client_wrapper)
        self.document = AsyncDocumentClient(client_wrapper=self._client_wrapper)
        self.code_execution = AsyncCodeExecutionClient(client_wrapper=self._client_wrapper)

    async def trigger_error_sentry_debug_get(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.trigger_error_sentry_debug_get()


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "sentry-debug",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def health_check_health_check_get(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.health_check_health_check_get()


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "health_check",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


def _get_base_url(*, base_url: typing.Optional[str] = None, environment: AxiomaticEnvironment) -> str:
    if base_url is not None:
        return base_url
    elif environment is not None:
        return environment.value
    else:
        raise Exception("Please pass in either base_url or environment to construct the client")



# This file was auto-generated by Fern from our API Definition.

import typing

StatementType = typing.Union[
    typing.Literal["COST_FUNCTION", "STRUCTURE_CONSTRAINT", "PARAMETER_CONSTRAINT", "QUERY_SUMMARIZER", "UNKNOWN"],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

from .constraint_type import ConstraintType
from .execute_code_response import ExecuteCodeResponse
from .extract_text_response import ExtractTextResponse
from .find_mappings_response import FindMappingsResponse
from .formalize_circuit_response import FormalizeCircuitResponse
from .formalize_response import FormalizeResponse
from .generate_code_response import GenerateCodeResponse
from .generate_component_code_response import GenerateComponentCodeResponse
from .http_validation_error import HttpValidationError
from .interactive_response import InteractiveResponse
from .mapping import Mapping
from .mapping_computation_arguments_value import MappingComputationArgumentsValue
from .netlist import Netlist
from .optimize_netlist_response import OptimizeNetlistResponse
from .parse_response import ParseResponse
from .pic_component import PicComponent
from .pic_component_info_value import PicComponentInfoValue
from .pic_component_settings_value import PicComponentSettingsValue
from .refine_code_response import RefineCodeResponse
from .refine_component_code_response import RefineComponentCodeResponse
from .requirement_body import RequirementBody
from .solution_response import SolutionResponse
from .solution_response_solution_value import SolutionResponseSolutionValue
from .statement import Statement
from .statement_type import StatementType
from .validate_response import ValidateResponse
from .validation_error import ValidationError
from .validation_error_loc_item import ValidationErrorLocItem
from .verify_circuit_code_response import VerifyCircuitCodeResponse
from .verify_response import VerifyResponse

__all__ = [
    "ConstraintType",
    "ExecuteCodeResponse",
    "ExtractTextResponse",
    "FindMappingsResponse",
    "FormalizeCircuitResponse",
    "FormalizeResponse",
    "GenerateCodeResponse",
    "GenerateComponentCodeResponse",
    "HttpValidationError",
    "InteractiveResponse",
    "Mapping",
    "MappingComputationArgumentsValue",
    "Netlist",
    "OptimizeNetlistResponse",
    "ParseResponse",
    "PicComponent",
    "PicComponentInfoValue",
    "PicComponentSettingsValue",
    "RefineCodeResponse",
    "RefineComponentCodeResponse",
    "RequirementBody",
    "SolutionResponse",
    "SolutionResponseSolutionValue",
    "Statement",
    "StatementType",
    "ValidateResponse",
    "ValidationError",
    "ValidationErrorLocItem",
    "VerifyCircuitCodeResponse",
    "VerifyResponse",
]



# This file was auto-generated by Fern from our API Definition.

from ..core.client_wrapper import SyncClientWrapper
import typing
from ..core.request_options import RequestOptions
from ..types.interactive_response import InteractiveResponse
from ..core.pydantic_utilities import parse_obj_as
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper


class ExperimentalClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def assistant(
        self, *, query: str, context: str, request_options: typing.Optional[RequestOptions] = None
    ) -> InteractiveResponse:
        """
        Interactive assistant for IDE extension

        Parameters
        ----------
        query : str

        context : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        InteractiveResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.experimental.assistant(
            query="query",
            context="context",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "experimental/assistant",
            method="POST",
            params={
                "query": query,
                "context": context,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    InteractiveResponse,
                    parse_obj_as(
                        type_=InteractiveResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncExperimentalClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def assistant(
        self, *, query: str, context: str, request_options: typing.Optional[RequestOptions] = None
    ) -> InteractiveResponse:
        """
        Interactive assistant for IDE extension

        Parameters
        ----------
        query : str

        context : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        InteractiveResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.experimental.assistant(
                query="query",
                context="context",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "experimental/assistant",
            method="POST",
            params={
                "query": query,
                "context": context,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    InteractiveResponse,
                    parse_obj_as(
                        type_=InteractiveResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



from IPython import get_ipython  # type: ignore
from IPython.core.magic import register_line_cell_magic, register_line_magic  # type: ignore
from IPython.display import HTML, display  # type: ignore

import platformdirs  # type: ignore
import os
import sys
from . import Axiomatic


class AXMagic:
    """Class implementing magic functions for IPython.
    Import with `%load_ext axiomatic.magic`."""

    client: Axiomatic
    query: str

    def __init__(self):
        self.folder = platformdirs.user_config_dir("axiomatic")

        self.client = Axiomatic()
        self.query = ""

    def ax_api(self, *args, **kwargs):
        pass

    def axquery(self, query, cell=None):
        if cell:
            # REFINE
            try:
                exec(cell)
                feedback = "Code executed successfully."
            except Exception as e:
                feedback = f"Errors:\n{e}"
            print(feedback)
            current_query = f"{self.query}\n{query}"
            result = self.client.pic.circuit.refine(
                query=current_query, code=cell, feedback=feedback
            )
        else:
            # GENERATE FROM SCRATCH
            self.query = query
            result = self.client.pic.circuit.generate(query=query)

        # Process output
        pre_thought = result.raw_content.split("<thought>")[0]
        thought = result.thought_text.replace("\n", "<br>")
        if not thought:
            output = result.raw_content.replace("\n", "<br>")
        else:
            output = pre_thought + "<br>" + thought
        html_content = f"""<div style='font-family: Arial, sans-serif; line-height: 1.5;'>"
<div style='color: #6EB700;'><strong>AX:</strong> {output}</div>"""
        display(HTML(html_content))

        # Process code
        # remove last three lines (saving file)
        if result.code:
            code = "\n".join(result.code.split("\n")[:-3] + ["c"])
            if "google.colab" in sys.modules:
                # When running in colab
                from google.colab import _frontend  # type: ignore

                _frontend.create_scratch_cell(
                    f"""# {query}\n# %%ax_fix\n{code}""", bottom_pane=True
                )
            else:
                # When running in jupyter
                get_ipython().set_next_input(f"# %%ax_fix\n{code}", replace=False)


    def ax_query(self, query, cell=None):
        # Updates the target circuit query
        self.query = query
        return self.axquery(query, cell)

    def ax_fix(self, query, cell=None):
        # Runs again without updating the query
        return self.axquery(query, cell)


def ax_help(value: str):
    print(
        """
Available commands:

- `%load_ext axiomatic_pic` loads the ipython extension.
- `%ax_query` returns the requested circuit using our experimental API
- `%%ax_fix` edit the given code
"""
    )


def load_ipython_extension(ipython):
    ax_magic = AXMagic()
    ipython.register_magic_function(ax_magic.ax_query, "line_cell")
    ipython.register_magic_function(ax_magic.ax_fix, "line_cell")
    ipython.register_magic_function(ax_magic.ax_api, "line")
    ipython.register_magic_function(ax_help, "line")



# This file was auto-generated by Fern from our API Definition.

from .types import (
    ConstraintType,
    ExecuteCodeResponse,
    ExtractTextResponse,
    FindMappingsResponse,
    FormalizeCircuitResponse,
    FormalizeResponse,
    GenerateCodeResponse,
    GenerateComponentCodeResponse,
    HttpValidationError,
    InteractiveResponse,
    Mapping,
    MappingComputationArgumentsValue,
    Netlist,
    OptimizeNetlistResponse,
    ParseResponse,
    PicComponent,
    PicComponentInfoValue,
    PicComponentSettingsValue,
    RefineCodeResponse,
    RefineComponentCodeResponse,
    RequirementBody,
    SolutionResponse,
    SolutionResponseSolutionValue,
    Statement,
    StatementType,
    ValidateResponse,
    ValidationError,
    ValidationErrorLocItem,
    VerifyCircuitCodeResponse,
    VerifyResponse,
)
from .errors import UnprocessableEntityError
from . import code_execution, document, experimental, formalization, generic, lean, pic, requirements
from .client import AsyncAxiomatic, Axiomatic
from .environment import AxiomaticEnvironment
from .version import __version__

__all__ = [
    "AsyncAxiomatic",
    "Axiomatic",
    "AxiomaticEnvironment",
    "ConstraintType",
    "ExecuteCodeResponse",
    "ExtractTextResponse",
    "FindMappingsResponse",
    "FormalizeCircuitResponse",
    "FormalizeResponse",
    "GenerateCodeResponse",
    "GenerateComponentCodeResponse",
    "HttpValidationError",
    "InteractiveResponse",
    "Mapping",
    "MappingComputationArgumentsValue",
    "Netlist",
    "OptimizeNetlistResponse",
    "ParseResponse",
    "PicComponent",
    "PicComponentInfoValue",
    "PicComponentSettingsValue",
    "RefineCodeResponse",
    "RefineComponentCodeResponse",
    "RequirementBody",
    "SolutionResponse",
    "SolutionResponseSolutionValue",
    "Statement",
    "StatementType",
    "UnprocessableEntityError",
    "ValidateResponse",
    "ValidationError",
    "ValidationErrorLocItem",
    "VerifyCircuitCodeResponse",
    "VerifyResponse",
    "__version__",
    "code_execution",
    "document",
    "experimental",
    "formalization",
    "generic",
    "lean",
    "pic",
    "requirements",
]



# This file was auto-generated by Fern from our API Definition.

import typing
from ..core.client_wrapper import SyncClientWrapper
from .circuit.client import CircuitClient
from .component.client import ComponentClient
from ..types.statement import Statement
from ..types.netlist import Netlist
from ..core.request_options import RequestOptions
from ..types.find_mappings_response import FindMappingsResponse
from ..core.serialization import convert_and_respect_annotation_metadata
from ..core.pydantic_utilities import parse_obj_as
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper
from .circuit.client import AsyncCircuitClient
from .component.client import AsyncComponentClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class PicClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.circuit = CircuitClient(client_wrapper=self._client_wrapper)
        self.component = ComponentClient(client_wrapper=self._client_wrapper)

    def find_mappings(
        self,
        *,
        statements: typing.Sequence[Statement],
        netlist: Netlist,
        max_iter: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FindMappingsResponse:
        """
        Map variables in the constraints to computations on the netlist.

        Parameters
        ----------
        statements : typing.Sequence[Statement]

        netlist : Netlist

        max_iter : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FindMappingsResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic, Netlist, PicComponent, Statement

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.pic.find_mappings(
            statements=[
                Statement(
                    id="id",
                    statement="statement",
                )
            ],
            netlist=Netlist(
                name="name",
                instances={
                    "key": PicComponent(
                        component="component",
                    )
                },
                connections={"key": "value"},
                ports={"key": "value"},
            ),
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "pic/circuit/mappings/find",
            method="POST",
            params={
                "max_iter": max_iter,
            },
            json={
                "statements": convert_and_respect_annotation_metadata(
                    object_=statements, annotation=typing.Sequence[Statement], direction="write"
                ),
                "netlist": convert_and_respect_annotation_metadata(
                    object_=netlist, annotation=Netlist, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    FindMappingsResponse,
                    parse_obj_as(
                        type_=FindMappingsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncPicClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.circuit = AsyncCircuitClient(client_wrapper=self._client_wrapper)
        self.component = AsyncComponentClient(client_wrapper=self._client_wrapper)

    async def find_mappings(
        self,
        *,
        statements: typing.Sequence[Statement],
        netlist: Netlist,
        max_iter: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FindMappingsResponse:
        """
        Map variables in the constraints to computations on the netlist.

        Parameters
        ----------
        statements : typing.Sequence[Statement]

        netlist : Netlist

        max_iter : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FindMappingsResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic, Netlist, PicComponent, Statement

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.pic.find_mappings(
                statements=[
                    Statement(
                        id="id",
                        statement="statement",
                    )
                ],
                netlist=Netlist(
                    name="name",
                    instances={
                        "key": PicComponent(
                            component="component",
                        )
                    },
                    connections={"key": "value"},
                    ports={"key": "value"},
                ),
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "pic/circuit/mappings/find",
            method="POST",
            params={
                "max_iter": max_iter,
            },
            json={
                "statements": convert_and_respect_annotation_metadata(
                    object_=statements, annotation=typing.Sequence[Statement], direction="write"
                ),
                "netlist": convert_and_respect_annotation_metadata(
                    object_=netlist, annotation=Netlist, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    FindMappingsResponse,
                    parse_obj_as(
                        type_=FindMappingsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)
